using HarmonyLib;
using System;
using System.Text;
using System.Collections.Generic;
using System.Reflection;
using UnityEngine;

public class ConfigMachines : Mod
{
    public static JsonModInfo modInfo;
    public static List<ValueTuple<SettingType, string>> modSettings = new List<ValueTuple<SettingType, string>>();
    public static Settings ExtraSettingsAPI_Settings;

    private Harmony harmony;
    private static bool debugging = false;
    private static Dictionary<CookingSlot, ValueTuple<string, float>> cookingSlotMults = new Dictionary<CookingSlot, ValueTuple<string, float>>();

    public void Start()
    {
        ConfigMachines.modInfo = modlistEntry.jsonmodinfo;
        this.harmony = new Harmony("com.destruction.ConfigurableMachines");
        this.harmony.PatchAll(Assembly.GetExecutingAssembly());
        ConfigMachines.ExtraSettingsAPI_Settings = new Settings();

        // Place a default instance of the Settings class in our variable.
        // Extra settings api should override this anyways.

        // Find all the slots and add them to our update list.
        foreach (var slot in FindObjectsOfType<CookingSlot>())
        {
            // Only add if they are awake.
            if (slot.gameObject.activeInHierarchy)
            {
                ConfigMachines.AddCookingSlot(slot);
            }
        }

        // Load the json data so we can iterate over settings.
        var jObject = new JSONObject(Encoding.Default.GetString(GetEmbeddedFileBytes("modinfo.json")));
        jObject = jObject.GetField("modSettings");
        foreach(var x in jObject.list)
        {
            var name = x.GetField("name").str;
            SettingType type = (SettingType)Enum.Parse(typeof(SettingType), x.GetField("type").str);
            ConfigMachines.modSettings.Add(ValueTuple.Create<SettingType, string>(type, name));
        }

        ConfigMachines.Log("Mod has been loaded.");
    }

    public void OnModUnload()
    {
        ConfigMachines.RestoreMultipliers();
        this.harmony.UnpatchAll("com.destruction.ConfigurableMachines");
        ConfigMachines.Log("Mod has been unloaded.");
    }

    public static void Log(object message)
    {
        Debug.Log("[" + modInfo.name + "]: " + message.ToString());
    }

    public static void DebugLog(object message)
    {
        Debug.Log("[" + modInfo.name + "][DEBUG]: " + message.ToString());
    }

    public static void AddCookingSlot(CookingSlot slot)
    {
        // Get the original multiplier.
        float originalMult = (float)Traverse.Create(slot).Field("cookTimeMultiplier").GetValue();

        string name = slot.gameObject.name;


        ValueTuple<string, float> valueTup = ("", originalMult);
        bool valueTupSet = false;

        // Now figure out where to put it.
        if (name.StartsWith("CookingSlot_Smelter_Electric"))
        {
            // Electric smelter slot.
            valueTup.Item1 = "electricSmelterMultiplier";
            valueTupSet = true;
        }
        else if (name.StartsWith("CookingSlot_Color"))
        {
            // Paint mill slot.
            valueTup.Item1 = "paintMillMultiplier";
            valueTupSet = true;
        }
        else if (name.StartsWith("CookingSlot_Food_BigGrill"))
        {
            if (slot.GetComponentInParent<CookingStand_Electric>())
            {
                // Electric grill slot.
                valueTup.Item1 = "electricGrillMultiplier";
                valueTupSet = true;
            }
            else
            {
                // Advanced grill slot.
                valueTup.Item1 = "advancedGrillMultiplier";
                valueTupSet = true;
            }
        }
        else if (name.StartsWith("CookingSlot_Food"))
        {
            // Basic grill slot.
            valueTup.Item1 = "basicGrillMultiplier";
            valueTupSet = true;
        }
        else if (name.StartsWith("CookingSlot_Smelter"))
        {
            // Basic smelter.
            valueTup.Item1 = "basicSmelterMultiplier";
            valueTupSet = true;
        }
        else if (slot.gameObject.GetComponentInParent<Block_CookingStand_Purifier>())
        {
            name = slot.gameObject.GetComponentInParent<Block_CookingStand_Purifier>().gameObject.name;
            if (name.StartsWith("Placeable_CookingStand_Purifier_One"))
            {
                // Basic purifier.
                valueTup.Item1 = "basicPurifierMultiplier";
                valueTupSet = true;
            }
            else if (name.StartsWith("Placeable_CookingStand_Purifier_Two"))
            {
                // Advanced purifier.
                valueTup.Item1 = "advancedPurifierMultiplier";
                valueTupSet = true;
            }
        }

        if (valueTupSet)
        {
            cookingSlotMults.TryAdd(slot, valueTup);
            ConfigMachines.UpdateCookingSlot(slot);
        }
        else if (ConfigMachines.debugging)
        {
            ConfigMachines.DebugLog("Found cooking slot with name \"" + name + "\" but could not identify it.");
        }
    }

    public static void RestoreMultipliers()
    {
        // Restore all multipliers to their original values.
        cookingSlotMults.Clean();
        foreach (var slot in cookingSlotMults.Keys)
        {
            Traverse.Create(slot).Field("cookTimeMultiplier").SetValue(cookingSlotMults[slot].Item2);
        }
    }

    public static void UpdateAllSlots()
    {
        if (debugging)
        {
            ConfigMachines.DebugLog("Updating all slots.");
        }

        ConfigMachines.cookingSlotMults.Clean();
        foreach(CookingSlot k in ConfigMachines.cookingSlotMults.Keys)
        {
            ConfigMachines.UpdateCookingSlot(k);
        }
    }

    private static void UpdateCookingSlot(CookingSlot slot)
    {
        try
        {
            var field = Traverse.Create(slot).Field("cookTimeMultiplier");
            // Get the tuple with the setting name and
            var valueTup = ConfigMachines.cookingSlotMults[slot];
            if (debugging)
            {
                ConfigMachines.DebugLog($"Fetched value of {ConfigMachines.ExtraSettingsAPI_Settings.GetSlider(valueTup.Item1.ToString())} for setting \"{valueTup.Item1.ToString()}\"");
            }
            var mult = ConfigMachines.ExtraSettingsAPI_Settings.GetSlider(valueTup.Item1.ToString());
            field.SetValue(mult * valueTup.Item2);

            if (debugging)
            {
                ConfigMachines.DebugLog("Updated CookingSlot \"" + slot.ToString() + "\" to new multiplier of " + (mult * valueTup.Item2).ToString() + ".");
            }
        }
        catch (Exception e)
        {
            Debug.LogError(e.ToString());
            throw e;
        }
    }

    // Debug console commands.
    [ConsoleCommand(name: "ConfigMachines_ToggleDebug", docs: "Toggles debug logging.")]
    public static void ToggleDebug()
    {
        ConfigMachines.debugging = !ConfigMachines.debugging;
        var enabled = ConfigMachines.debugging ? "enabled" : "disabled";
        ConfigMachines.DebugLog($"Debug messages have been {enabled}.");
    }

    [ConsoleCommand(name: "ConfigMachines_DumpSettings", docs: "Dumps a list of setting for debug testing.")]
    public static void DumpSettings()
    {
        string debugMessage = "\nExtra Settings Loaded: " + (Settings.ExtraSettingsAPI_Loaded ? "True" : "False") + "\n";
        debugMessage += "Aquired Settings Values:\n";

        foreach(var valueTup in ConfigMachines.modSettings)
        {
            string line = "";
            switch (valueTup.Item1)
            {
                case SettingType.slider:
                    line = $"    {valueTup.Item2}: {ConfigMachines.ExtraSettingsAPI_Settings.GetSlider(valueTup.Item2)}\n";
                    break;
            }
            debugMessage += line;
        }

        ConfigMachines.DebugLog(debugMessage);
    }

    [ConsoleCommand(name: "ConfigMachines_DumpAll", docs: "Creates an extensive dump of debugging information.")]
    public static void DumpAll()
    {
        // Start with the settings.
        string debugMessage = "\nExtra Settings Loaded: " + (Settings.ExtraSettingsAPI_Loaded ? "True" : "False") + "\n";
        debugMessage += "Aquired Settings Values:\n";

        foreach(var valueTup in ConfigMachines.modSettings)
        {
            string line = "";
            switch (valueTup.Item1)
            {
                case SettingType.slider:
                    line = $"    {valueTup.Item2}: {ConfigMachines.ExtraSettingsAPI_Settings.GetSlider(valueTup.Item2)}\n";
                    break;
            }
            debugMessage += line;
        }

        // Now we want to look at the saved cooking slots.
        debugMessage += "\nSaved Cooking Slot States:\n";
        foreach (CookingSlot slot in ConfigMachines.cookingSlotMults.Keys)
        {
            debugMessage += "    Slot " + slot.ToString() + ":\n";
            if (slot)
            {
                // We only want to get a bunch of data if the object is good.
                // Get the saved data into short variables.
                ValueTuple<string, float> valueTup = ConfigMachines.cookingSlotMults[slot];
                var currentMult = Traverse.Create(slot).Field("cookTimeMultiplier").GetValue();
                debugMessage += "        Is Valid: True\n";
                debugMessage += $"        Original Mutiplier: {valueTup.Item2}\n";
                debugMessage += $"        Current Multiplier: {currentMult}\n";
                debugMessage += $"        Setting name: {valueTup.Item1}\n";
            }
            else
            {
                debugMessage += "        Is Valid: False\n";
            }
        }

        ConfigMachines.DebugLog(debugMessage);
    }

    [HarmonyPatch(typeof(CookingSlot), "Awake")]
    public class Patch_CookingSlotAwake
    {
        static void Postfix(ref CookingSlot __instance)
        {
            // On awake, add to the dictionary. This will update it.
            ConfigMachines.AddCookingSlot(__instance);
        }
    }
}



public class Settings
{
    public float GetSlider(string name)
    {
        return ExtraSettingsAPI_GetSliderValue(name);
    }

    public void ExtraSettingsAPI_Load() // Occurs when the API loads the mod's settings
    {
        ConfigMachines.ExtraSettingsAPI_Settings = this;
        ConfigMachines.UpdateAllSlots();
    }

    public void ExtraSettingsAPI_SettingsClose() // Occurs when the API loads the mod's settings
    {
        ConfigMachines.UpdateAllSlots();
    }

    public static float ExtraSettingsAPI_GetSliderValue(string SettingName) => 1f;
    public static bool ExtraSettingsAPI_Loaded = false;
}



public enum SettingType
{
    checkbox,
    slider,
    combobox,
    keybind,
    button,
    text,
    data,
    input
}



static class DictionaryExtension
{
    // Thanks Aidan.
    public static void Clean<X,Y>(this Dictionary<X,Y> c) where X : MonoBehaviour
    {
        var l = new List<X>();
        foreach (var k in c.Keys)
        {
            if (!k)
            {
                l.Add(k);
            }
        }
        foreach (var i in l)
        {
            c.Remove(i);
        }
    }
}
